@@ -11,7 +11,7 @@ used within the context. Devices include theoretical sources and detectors.
 """
 
 from cmath import rect
-from typing import TYPE_CHECKING, ClassVar, List, Optional
+from typing import TYPE_CHECKING, ClassVar, List, Optional, Tuple
 
 import numpy as np
 from scipy.constants import epsilon_0, h, mu_0
@@ -31,6 +31,16 @@ context = None
 nprect = np.vectorize(rect)
 
 
+def from_db(x: float, factor: int = 10) -> float:
+    """Converts a dB value to a linear value."""
+    return 10 ** (x / factor)
+
+
+def to_db(x: float, factor: int = 10) -> float:
+    """Converts a linear value to a dB value."""
+    return factor * np.log10(x)
+
+
 class Simulation:
     """This class instantiates a simulation context.
 
@@ -286,17 +296,25 @@ class Simulation:
             be the theoretical value of the circuit. If more than one sample is
             taken, they will vary based on simulated noise.
         """
-        # we enforce an odd number of samples so filter implementation is easy
+        # filtering requires an odd number of samples.
+        # if they want an even number, we will add one and then remove a
+        # sample at the end
+        _num_samples = num_samples
         if num_samples % 2 == 0:
-            raise ValueError("`num_samples` must be an odd number.")
+            _num_samples += 1
 
         # if we are taking more than one sample, include noise
-        self.num_samples = num_samples
+        self.num_samples = _num_samples
         self.noise = self.num_samples > 1
 
         # sample the signals
         signals = self._get_signals()
-        return signals
+
+        # remove the extra sample if we added one
+        if _num_samples != num_samples:
+            return signals[:, :, :num_samples]
+        else:
+            return signals
 
     @classmethod
     def get_context(cls) -> "Simulation":
@@ -398,7 +416,7 @@ class Laser(Source):
         self._coupled_powers = np.array([])
         self._freqs = np.array([freq if freq else wl2freq(wl)])
         self._powers = np.array([power])
-        self.coupling_ratio = 10 ** (-np.abs(coupling_loss) / 10)
+        self.coupling_ratio = from_db(-np.abs(coupling_loss))
         self.freqs = np.array([])
         self.index = 0
         self.phase = phase
@@ -429,11 +447,7 @@ class Laser(Source):
         bandwidth :
             The bandwidth of the detector in Hz.
         """
-        return (
-            0
-            if self.rin == -np.inf
-            else 10 * np.log10((10 ** (self.rin / 10)) * bandwidth)
-        )
+        return 0 if self.rin == -np.inf else to_db(from_db(self.rin) * bandwidth)
 
     def get_rin_dist(self, i: int, j: int) -> List[float]:
         """Returns the normal distribution used for the i,j key. If this is the
@@ -544,12 +558,12 @@ class Detector(SimulationModel):
                     for source in self.context.sources:
                         # we let the laser handle the RIN distribution
                         # so the same noise is injected in all the signals
-                        rin = source.get_rin(self.high_fc - self.low_fc)
+                        rin = source.get_rin(self._get_bandwidth()[0])
                         dist = source.get_rin_dist(i, j)
 
                         # calculate the standard deviation of the RIN noise
                         std = (
-                            10 ** ((10 * np.log10(power[i][j][0]) + rin) / 20)
+                            from_db(to_db(power[i][j][0]) + rin, 20)
                             if power[i][j][0]
                             else 0
                         )
@@ -594,15 +608,15 @@ class Detector(SimulationModel):
         signal :
             The signal to filter.
         """
-        high = min(self.high_fc, 0.5 * (self.context.fs - 1))
+        bw = self._get_bandwidth()
         sos = (
-            butter(6, high, "lowpass", fs=self.context.fs, output="sos")
-            if self.low_fc == 0
+            butter(6, bw[2], "lowpass", fs=self.context.fs, output="sos")
+            if bw[1] == 0
             else butter(
                 6,
                 [
-                    max(self.low_fc, self.context.fs / self.context.num_samples * 30),
-                    high,
+                    bw[1],
+                    bw[2],
                 ],
                 "bandpass",
                 fs=self.context.fs,
@@ -611,6 +625,20 @@ class Detector(SimulationModel):
         )
         return sosfiltfilt(sos, signal)
 
+    def _get_bandwidth(self):
+        """Gets the bandwidth of the detector w.r.t.
+
+        the sampling frequency.
+        """
+        low = (
+            0
+            if self.low_fc == 0
+            else max(self.low_fc, self.context.fs / self.context.num_samples * 30)
+        )
+        high = min(self.high_fc, 0.5 * (self.context.fs - 1))
+
+        return (high - low, low, high)
+
 
 class DifferentialDetector(Detector):
     """A differential detector takes two connections and provides three outputs
@@ -689,27 +717,32 @@ class DifferentialDetector(Detector):
                     for source in self.context.sources:
                         # get the RIN specs from the laser to ensure that the
                         # same noise is injected across all signals
-                        monitor_rin = source.get_rin(
-                            self.monitor_high_fc - self.monitor_low_fc
-                        )
-                        rf_rin = source.get_rin(self.rf_high_fc - self.rf_low_fc)
+                        monitor_rin = source.get_rin(self._get_monitor_bandwidth()[0])
+                        rf_rin = source.get_rin(self._get_rf_bandwidth()[0])
                         dist = source.get_rin_dist(i, j)
 
+                        # if the two powers are different, we need to adjust
+                        # the CMRR to account for the difference
+                        cmrr = self.rf_cmrr
+                        sum = p1[i][j][0] + p2[i][j][0]
+                        diff = np.abs(p1[i][j][0] - p2[i][j][0])
+                        if diff:
+                            cmrr2 = -to_db(sum / diff)
+                            cmrr = to_db(
+                                np.sqrt(from_db(cmrr) ** 2 + from_db(cmrr2) ** 2)
+                            )
+
                         # only calculate the noise if there is power
                         if p1[i][j][0] > 0:
-                            p1db = 10 * np.log10(p1[i][j][0])
-                            monitor_noise1 += (10 ** ((p1db + monitor_rin) / 20)) * dist
-                            rf_noise1 += (
-                                10 ** ((p1db + rf_rin + self.rf_cmrr) / 20)
-                            ) * dist
+                            p1db = to_db(p1[i][j][0])
+                            monitor_noise1 += from_db(p1db + monitor_rin, 20) * dist
+                            rf_noise1 += from_db(p1db + rf_rin + cmrr, 20) * dist
 
                         # only calculate the noise if there is power
                         if p2[i][j][0] > 0:
-                            p2db = 10 * np.log10(p2[i][j][0])
-                            monitor_noise2 += (10 ** ((p2db + monitor_rin) / 20)) * dist
-                            rf_noise2 += (
-                                10 ** ((p2db + rf_rin + self.rf_cmrr) / 20)
-                            ) * dist
+                            p2db = to_db(p2[i][j][0])
+                            monitor_noise2 += from_db(p2db + monitor_rin, 20) * dist
+                            rf_noise2 += from_db(p2db + rf_rin + cmrr, 20) * dist
 
                     # store the RIN noise for later use
                     self.monitor_rin_dists1[i][j] = monitor_noise1
@@ -748,27 +781,24 @@ class DifferentialDetector(Detector):
             self._monitor(p2, self.monitor_rin_dists2),
         )
 
-    def _filter(self, signal: np.ndarray, low_fc: float, high_fc: float) -> np.ndarray:
+    def _filter(self, signal: np.ndarray, bw: Tuple[float, float, float]) -> np.ndarray:
         """Filters the signal.
 
         Parameters
         ----------
         signal :
             The signal to filter.
-        low_fc :
-            The lower cut-off frequency.
-        high_fc :
-            The higher cut-off frequency.
+        bw :
+            The bandwidth of the filter. (difference, low_fc, high_fc)
         """
-        high = min(high_fc, 0.5 * (self.context.fs - 1))
         sos = (
-            butter(6, high, "lowpass", fs=self.context.fs, output="sos")
-            if low_fc == 0
+            butter(6, bw[2], "lowpass", fs=self.context.fs, output="sos")
+            if bw[1] == 0
             else butter(
                 6,
                 [
-                    max(low_fc, self.context.fs / self.context.num_samples * 30),
-                    high,
+                    bw[1],
+                    bw[2],
                 ],
                 "bandpass",
                 fs=self.context.fs,
@@ -777,6 +807,39 @@ class DifferentialDetector(Detector):
         )
         return sosfiltfilt(sos, signal)
 
+    def _get_bandwidth(self, low_fc, high_fc):
+        """Gets the bandwidth of the detector w.r.t. the sampling frequency.
+
+        Parameters
+        ----------
+        low_fc :
+            The lower cut-off frequency.
+        high_fc :
+            The higher cut-off frequency.
+        """
+        low = (
+            0
+            if low_fc == 0
+            else max(low_fc, self.context.fs / self.context.num_samples * 30)
+        )
+        high = min(high_fc, 0.5 * (self.context.fs - 1))
+
+        return (high - low, low, high)
+
+    def _get_monitor_bandwidth(self):
+        """Gets the bandwidth of the monitor w.r.t.
+
+        the sampling frequency.
+        """
+        return self._get_bandwidth(self.monitor_low_fc, self.monitor_high_fc)
+
+    def _get_rf_bandwidth(self):
+        """Gets the bandwidth of the rf w.r.t.
+
+        the sampling frequency.
+        """
+        return self._get_bandwidth(self.rf_low_fc, self.rf_high_fc)
+
     def _monitor(self, power: np.ndarray, rin_dists: np.ndarray) -> np.ndarray:
         """Takes a signal and turns it into a monitor output.
 
@@ -805,7 +868,7 @@ class DifferentialDetector(Detector):
         signal :
             The signal to filter.
         """
-        return self._filter(signal, self.monitor_low_fc, self.monitor_high_fc)
+        return self._filter(signal, self._get_monitor_bandwidth())
 
     def _rf(self, p1: np.ndarray, p2: np.ndarray) -> np.ndarray:
         """Takes two signals and generates the differential RF signal. p1 - p2.
@@ -842,4 +905,4 @@ class DifferentialDetector(Detector):
         signal :
             The signal to filter.
         """
-        return self._filter(signal, self.rf_low_fc, self.rf_high_fc)
+        return self._filter(signal, self._get_rf_bandwidth())
